@@ -5,31 +5,9 @@ require 'forwardable'
 module Faraday
   # Response represents an HTTP response from making an HTTP request.
   class Response
-    # Used for simple response middleware.
-    class Middleware < Faraday::Middleware
-      def call(env)
-        @app.call(env).on_complete do |environment|
-          on_complete(environment)
-        end
-      end
-
-      # Override this to modify the environment after the response has finished.
-      # Calls the `parse` method if defined
-      # `parse` method can be defined as private, public and protected
-      def on_complete(env)
-        return unless respond_to?(:parse, true) && env.parse_body?
-
-        env.body = parse(env.body)
-      end
-    end
-
     extend Forwardable
     extend MiddlewareRegistry
 
-    register_middleware File.expand_path('response', __dir__),
-                        raise_error: [:RaiseError, 'raise_error'],
-                        logger: [:Logger, 'logger']
-
     def initialize(env = nil)
       @env = Env.from(env) if env
       @on_complete_callbacks = []
@@ -48,6 +26,7 @@ module Faraday
     def headers
       finished? ? env.response_headers : {}
     end
+
     def_delegator :headers, :[]
 
     def body
@@ -59,10 +38,10 @@ module Faraday
     end
 
     def on_complete(&block)
-      if !finished?
-        @on_complete_callbacks << block
-      else
+      if finished?
         yield(env)
+      else
+        @on_complete_callbacks << block
       end
       self
     end
@@ -105,3 +84,7 @@ module Faraday
     end
   end
 end
+
+require 'faraday/response/json'
+require 'faraday/response/logger'
+require 'faraday/response/raise_error'
